#
# Copyright 2014 Intel Corp
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from oslo_utils import netutils

from ceilometer.hardware import inspector
from ceilometer.tests import base


class TestHardwareInspector(base.BaseTestCase):
    def test_get_inspector(self):
        url = netutils.urlsplit("snmp://")
        driver = inspector.get_inspector(url)
        self.assertTrue(driver)

    def test_get_inspector_illegal(self):
        url = netutils.urlsplit("illegal://")
        self.assertRaises(RuntimeError,
                          inspector.get_inspector,
                          url)
